import logging
import json
import calendar
from datetime import timedelta, datetime
import yaml
import requests

from concurrent.futures import ThreadPoolExecutor, as_completed
from celery import shared_task, chain
from django.utils import timezone
from . import app_settings
from .helpers import OreHelper
from .models import MiningObservation, MoonFrack, FrackOre, MoonRental, OrePrice, OreTaxRates, OreTax, InvoiceRecord
from corptools.models import MapSystem, Notification, EveLocation, MapSystemMoon, EveItemType, CorporationAudit, EveName
from corptools import providers
from corptools.task_helpers.corp_helpers import get_corp_token
from corptools.task_helpers.update_tasks import fetch_location_name
from allianceauth.services.tasks import QueueOnce
from .app_settings import PUBLIC_MOON_CORPS

logger = logging.getLogger(__name__)


def filetime_to_dt(ft):
    us = (ft - 116444736000000000) // 10
    return datetime(1970, 1, 1) + timedelta(microseconds=us)


@shared_task
def cleanup_all_fracks():
    for f in MoonFrack.objects.all().values_list("structure", flat=True).distinct():
        cleanup_moon_pulls.delay(f)


@shared_task
def cleanup_moon_pulls(observer_id):
    fracks = list(MoonFrack.objects.filter(
        structure_id=observer_id).order_by('arrival_time'))
    removals = []
    for index, frack in enumerate(fracks):
        if index+1 < len(fracks):
            if frack.arrival_time > fracks[index + 1].start_time:
                removals.append(frack)
    cnt = len(removals)
    if cnt:
        for df in removals:
            df.delete()
        return f"Removed {cnt} bad extractions."
    return f"No bad extractions."


@shared_task
def process_moon_pulls():
    logger.debug("Started Mining Pull Sync")
    start_time = timezone.now() - timedelta(days=90)

    notification_type_filter = ['MoonminingExtractionStarted', ]

    # YAML Structure
    #    autoTime: #####
    #    moonID: ####
    #    oreVolumeByType:
    #        45490: 3981232.1291307067
    #        45499: 7115599.461194293
    #        45504: 4589835.07632598
    #    readyTime: #####
    #    solarSystemID: ####
    #    startedBy: ####
    #    startedByLink: <a href="showinfo:1385//###">CCCC</a>
    #    structureID: ####
    #    structureLink: <a href="showinfo:35835//####">CCCC</a>
    #    structureName: CCCCCCCCC
    #    structureTypeID: ####

    notifications = Notification.objects.filter(character__character__corporation_id__in=PUBLIC_MOON_CORPS,
                                                timestamp__gte=start_time,
                                                notification_type__in=notification_type_filter).select_related('notification_text')

    fracks = MoonFrack.objects.filter(start_time__gte=start_time)
    notification_ids = set(fracks.values_list(
        'notification__notification_id', flat=True))
    frack_dict = {}
    for frack in fracks:
        if frack.moon_id not in frack_dict:
            frack_dict[frack.moon_id] = {}
        frack_dict[frack.moon_id][frack.start_time] = frack

    _type_list = []
    _ores = []
    for notification in notifications:
        try:
            if notification.notification_id not in notification_ids:
                notification_ids.add(notification.notification_id)
                notification_data = yaml.load(
                    notification.notification_text.notification_text, Loader=yaml.UnsafeLoader)
                moon_id = notification_data['moonID']
                start_time = notification.timestamp

                new_frack = False

                if moon_id in frack_dict:
                    if start_time not in frack_dict[moon_id]:
                        new_frack = True
                else:
                    new_frack = True

                if new_frack:
                    _structure, _c = EveLocation.objects.update_or_create(location_id=notification_data['structureID'],
                                                                          defaults={
                        "location_name": notification_data['structureName'],
                        "system_id": notification_data['solarSystemID']
                    })
                    _moon, _c = MapSystemMoon.objects.get_or_create_from_esi(
                        moon_id=moon_id)
                    _corp_audit = CorporationAudit.objects.get(
                        corporation__corporation_id=notification.character.character.corporation_id)
                    _frack = MoonFrack.objects.create(corporation=_corp_audit,
                                                      moon_name=_moon,
                                                      moon_id=moon_id,
                                                      notification=notification,
                                                      structure=_structure,
                                                      start_time=start_time,
                                                      arrival_time=filetime_to_dt(
                                                          notification_data['readyTime']),
                                                      auto_time=filetime_to_dt(
                                                          notification_data['autoTime'])
                                                      )

                    for ore, volume in notification_data['oreVolumeByType'].items():
                        if ore not in _type_list:
                            _type_list.append(ore)
                        _ores.append(FrackOre(frack=_frack,
                                              ore_id=ore,
                                              total_m3=volume
                                              ))
        except Exception as e:
            logger.warn("Failed to process moon", exc_info=1)
            pass

    EveItemType.objects.create_bulk_from_esi(_type_list)
    FrackOre.objects.bulk_create(_ores, batch_size=500)

    return "fetched!"


@shared_task
def run_obs_for_all_corps(force=False):
    for c in CorporationAudit.objects.all():
        logger.debug(f"Sending Obs update for {c.corporation}")
        queue_moon_obs.apply_async(
            args=[c.corporation.corporation_id], priority=6)


@shared_task
def queue_moon_obs(corp_id, force=False):
    logger.debug("Started Mining Ob Sync for {}".format(corp_id))

    token = get_corp_token(
        corp_id, ['esi-industry.read_corporation_mining.v1'], ['Accountant', 'Director'])
    obs = providers.esi.client.Industry.get_corporation_corporation_id_mining_observers(corporation_id=corp_id,
                                                                                        token=token.valid_access_token()).results()
    task_queue = []
    for ob in obs:
        task_queue.append(process_moon_obs.si(ob.get('observer_id'), corp_id))
    chain(task_queue).apply_async(priority=8)


@shared_task(bind=True, base=QueueOnce)
def process_moon_obs(self, observer_id, corporation_id):
    logger.debug("Started Mining Ob Sync for {}".format(observer_id))

    token = get_corp_token(corporation_id, [
                           'esi-industry.read_corporation_mining.v1', 'esi-universe.read_structures.v1'], ['Accountant', 'Director'])
    obs = providers.esi.client.Industry.get_corporation_corporation_id_mining_observers_observer_id(corporation_id=corporation_id,
                                                                                                    observer_id=observer_id,
                                                                                                    token=token.valid_access_token()).results()
    eve_names = set(EveName.objects.all().values_list('eve_id', flat=True))
    type_names = set(EveItemType.objects.all(
    ).values_list('type_id', flat=True))
    corp = CorporationAudit.objects.get(
        corporation__corporation_id=corporation_id)

    observer = None
    structure_exists = False
    moon = None

    try:
        observer = EveLocation.objects.get(location_id=observer_id)
        structure_exists = True

    except EveLocation.DoesNotExist:
        _ob = fetch_location_name(
            observer_id, "solar_system", token.character_id)
        if _ob:
            _ob.save()
            observer = _ob
            structure_exists = True

    _moon = MoonFrack.objects.filter(
        structure_id=observer_id, moon_name__isnull=False)
    if _moon.exists():
        moon = _moon.last().moon_name

    ob_pks = set(MiningObservation.objects.filter(
        observing_id=observer_id).values_list('ob_pk', flat=True))
    mining_ob_updates = []
    mining_ob_creates = []
    new_eve_names = []
    new_item_types = []
    new_ob_pks = set()
    for ob in obs:
        pk = MiningObservation.build_pk(corporation_id, observer_id, ob.get(
            'character_id'), ob.get('last_updated'), ob.get('type_id'))

        if ob.get('character_id') not in eve_names:
            new_eve_names.append(ob.get('character_id'))
            eve_names.add(ob.get('character_id'))

        if ob.get('type_id') not in type_names:
            new_item_types.append(ob.get('type_id'))
            type_names.add(ob.get('type_id'))

        _ob = MiningObservation(ob_pk=pk,
                                observing_id=observer_id,
                                observing_corporation=corp,
                                character_name_id=ob.get('character_id'),
                                character_id=ob.get('character_id'),
                                last_updated=ob.get('last_updated'),
                                quantity=ob.get('quantity'),
                                recorded_corporation_id=ob.get(
                                    'recorded_corporation_id'),
                                type_id=ob.get('type_id'),
                                type_name_id=ob.get('type_id'),
                                moon=moon)

        if structure_exists:
            _ob.structure = observer

        if pk in ob_pks:
            mining_ob_updates.append(_ob)
        else:
            if pk in new_ob_pks:
                for _test_ob in mining_ob_creates:
                    if _test_ob.pk == pk:
                        _test_ob.quantity += _ob.quantity
                        break
            else:
                mining_ob_creates.append(_ob)
                new_ob_pks.add(pk)

    EveName.objects.create_bulk_from_esi(new_eve_names)
    EveItemType.objects.create_bulk_from_esi(new_item_types)

    if len(mining_ob_creates) > 0:
        MiningObservation.objects.bulk_create(mining_ob_creates)

    if len(mining_ob_updates) > 0:
        MiningObservation.objects.bulk_update(mining_ob_updates, [
                                              'quantity', 'last_updated', 'structure', 'observing_corporation'])

    msg = f"Corp:{corporation_id} Moon:{observer_id} Updated:{len(mining_ob_updates)} Created:{len(mining_ob_creates)}"

    corp.last_update_observers = timezone.now()
    corp.save()

    logger.debug(msg)
    return msg


@shared_task
def update_ore_prices():
    mins = OreHelper.get_mineral_array()
    minstr = str(mins.pop())
    for i in mins:
        minstr += f",{i}"
    url = f"https://market.fuzzwork.co.uk/aggregates/?station=60003760&types={minstr}"
    response = requests.get(url)
    price_data = response.json()
    price_cache = {}
    for key, item in price_data.items():
        name, created = EveItemType.objects.get_or_create_from_esi(key)
        name = name.name
        if name not in price_cache:
            price_cache[name] = {}
        price_cache[name]['the_forge'] = float(item['buy']['weightedAverage'])

    ores = OreHelper.get_ore_array()

    price_source = 'the_forge'  # We only use this now.

    for ore, minerals in ores.items():
        price = 0
        for mineral, qty in minerals["minerals"].items():
            price = price + (qty * price_cache[mineral][price_source])

        OrePrice.objects.update_or_create(item=minerals['model'],
                                          defaults={
            "price": price/minerals['portion']
        })

    update_tax_prices.apply_async(priority=7)

    return json.dumps(price_cache)


@shared_task
def update_tax_prices():
    taxs = OreTaxRates.objects.all()

    ores = OreHelper.get_ore_array_with_value()

    for tax in taxs:
        for id, o in ores.items():
            rate = getattr(tax, o['rarity'])
            price = float(o['value']) * (float(rate)/100) * \
                (float(tax.refine_rate)/100)
            OreTax.objects.update_or_create(
                item=o['model'],
                tax=tax,
                defaults={
                    "price": price
                }
            )


@shared_task
def generate_taxes():
    taxes = InvoiceRecord.generate_invoices()
    return f"Taxes Generated {taxes.base_ref} Total Mined:{taxes.total_mined:,} Total Tax:{taxes.total_taxed:,}"


def _get_system_planet_moons(system_id):
    system = providers.esi.client.Universe.get_universe_systems_system_id(
        system_id=system_id).result()
    out = []
    if system.get('planets', False):
        for p in system.get('planets'):
            if p.get('moons'):
                out += p.get('moons', [])
    return out


@shared_task
def process_moons_from_esi(regions):

    _moons = []
    _moon_models_updates = []
    _moon_models_creates = []

    _processes = []
    _current_systems = MapSystem.objects.filter(
        constellation__region_id__in=regions).values_list('system_id', flat=True)

    status = providers.esi.client.Status.get_status().result()

    with ThreadPoolExecutor(max_workers=50) as executor:
        for system in _current_systems:
            _processes.append(executor.submit(
                _get_system_planet_moons, system))

    for task in as_completed(_processes):
        _moons += task.result()

    _processes = []
    with ThreadPoolExecutor(max_workers=50) as executor:
        for moon in _moons:
            _processes.append(executor.submit(
                providers.esi._get_moon, moon, False))

    _moon_ids = set(
        list(MapSystemMoon.objects.all().values_list('moon_id', flat=True)))
    for task in as_completed(_processes):
        _moon_ob = task.result()
        if _moon_ob.moon_id in _moon_ids:
            _moon_models_updates.append(_moon_ob)
        else:
            _moon_models_creates.append(_moon_ob)

    if len(_moon_models_creates) > 0:
        MapSystemMoon.objects.bulk_create(
            _moon_models_creates, batch_size=1000)

    if len(_moon_models_updates) > 0:
        MapSystemMoon.objects.bulk_update(
            _moon_models_updates, ['name', 'x', 'y', 'z'], batch_size=1000)

    # sample_mem()
    output = "Moons: (Updated:{}, Created:{}) " \
             .format(len(_moon_models_updates),
                     len(_moon_models_creates))
    # memdump()
    return output


@shared_task
def invoice_moons():
    MoonRental.generate_invoices()


@shared_task
def invoice_single_moon(mrid):
    rental = MoonRental.objects.get(id=mrid)
    rng = calendar.monthrange(rental.start_date.year, rental.start_date.month)
    last_day = datetime(rental.start_date.year,
                        rental.start_date.month, rng[1]).day
    partial_price = round(rental.price / last_day *
                          (last_day-rental.start_date.day), -6)
    if partial_price > 30000000:
        due = timezone.now() + timedelta(days=14)
        inv = MoonRental.generate_invoice(
            rental.contact.id, [rental.moon.name], partial_price, due, single=True)
        inv.save()
        MoonRental.ping_invoice(inv)



from allianceauth.services.hooks import MenuItemHook, UrlHook
from django.utils.translation import gettext_lazy as _
from allianceauth import hooks
from . import urls


class MoonsBetaHook(MenuItemHook):
    def __init__(self):
        MenuItemHook.__init__(self,
                              _('Moon Board'),
                              'fas fa-moon fa-fw',
                              'moons:r',
                              navactive=['moons:r'])

    def render(self, request):
        if (request.user.has_perm('moons.view_available') or
            request.user.has_perm('moons.view_corp') or
            request.user.has_perm('moons.view_alliance') or
                request.user.has_perm('moons.view_all')):
            return MenuItemHook.render(self, request)
        return ''


@hooks.register('menu_item_hook')
def register_beta_menu():
    return MoonsBetaHook()


@hooks.register('url_hook')
def register_url():
    return UrlHook(urls, 'moons', r'^m/')


@hooks.register('discord_cogs_hook')
def register_cogs():
    return ["moons.mooncogs"]



# Generated by Django 3.2.8 on 2021-12-21 11:22

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('invoices', '0005_auto_20201013_0310'),
        ('corptools', '0059_invtypematerials_met_type'),
        ('eveonline', '0014_auto_20210105_1413'),
        ('moons', '0016_invoicerecord_base_ref'),
    ]

    operations = [
        migrations.CreateModel(
            name='MoonRental',
            fields=[
                ('id', models.AutoField(auto_created=True,
                 primary_key=True, serialize=False, verbose_name='ID')),
                ('note', models.TextField()),
                ('price', models.IntegerField(default=100000000)),
                ('start_date', models.DateTimeField()),
                ('end_date', models.DateTimeField(
                    blank=True, default=None, null=True)),
                ('contact', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE, to='eveonline.evecharacter')),
                ('corporation', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE, to='eveonline.evecorporationinfo')),
                ('last_invoice', models.ForeignKey(blank=True, default=None, null=True,
                 on_delete=django.db.models.deletion.SET_NULL, to='invoices.invoice')),
                ('moon', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE, to='corptools.mapsystemmoon')),
            ],
        ),
    ]



from django.urls import re_path
from .api import api

from . import views
app_name = 'moons'

urlpatterns = [
    re_path(r'^$', views.extractions, name='list'),
    re_path(r'^data/', views.observers, name='data'),
    re_path(r'^r/', views.react, name='r'),
    re_path(r'^api/', api.urls),

]



from allianceauth.eveonline.models import EveCharacter
from datetime import timedelta

from typing import List
from django.utils import timezone
from django.utils.timezone import activate
from moons.helpers import OreHelper, what_frack_id

from ninja import NinjaAPI, Form, main
from ninja.security import django_auth
from ninja.responses import codes_4xx

from django.core.exceptions import PermissionDenied
from django.db.models import F, Sum, Q, Max, Min
from django.db.models import Subquery, OuterRef
from django.db.models import FloatField, F, ExpressionWrapper

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from django.conf import settings
from esi.models import Token

from . import models
from corptools.models import CharacterAudit, CorporationAudit, MapSystemMoon
from . import schema

from invoices.models import Invoice

from django.db.models import Sum

import logging

logger = logging.getLogger(__name__)


api = NinjaAPI(title="MoonTool API", version="0.0.1",
               urls_namespace='moons:api', auth=django_auth, csrf=True,
               openapi_url=settings.DEBUG and "/openapi.json" or "")


@api.get(
    "/user/permissions",
    response={200: schema.MoonPermisions},
    tags=["User"]
)
def get_user_permisions(request):
    return {
        "view_public_extractions": request.user.has_perm('moons.view_available'),
        "view_corp_extractions": request.user.has_perm('moons.view_corp'),
        "view_alliance_extractions": request.user.has_perm('moons.view_alliance'),
        "view_observations": request.user.has_perm('moons.view_all'),
        "view_rentals": request.user.has_perm('moons.view_moonrental'),
        "edit_rentals": request.user.has_perm('moons.change_moonrental'),
        "su": request.user.is_superuser
    }


JACKPOT_IDS = [
    46281,  # Glistening Zeolites
    46283,  # Glistening Sylvite
    46285,  # Glistening Bitumens
    46287,  # Glistening Coesite
    46305,  # Glowing Carnotite
    46307,  # Glowing Zircon
    46309,  # Glowing Pollucite
    46311,  # Glowing Cinnabar
    46297,  # Shimmering Otavite
    46299,  # Shimmering Sperrylite
    46301,  # Shimmering Vanadinite
    46303,  # Shimmering Chromite
    46289,  # Twinkling Cobaltite
    46291,  # Twinkling Euxenite
    46293,  # Twinkling Titanite
    46295,  # Twinkling Scheelite
    46313,  # Shining Xenotime
    46315,  # Shining Monazite
    46317,  # Shining Loparite
    46319,  # Shining Ytterbite
]


@api.get(
    "/extractions/",
    response={200: List[schema.ExtractionEvent]},
    tags=["Observers"]
)
def get_moons_and_obs(request):
    if not request.user.has_perm("moons.view_available"):
        return []
    past_days = 3

    return get_moons_and_extractions(request, past_days)


"""
    start_date = timezone.now() - timedelta(days=past_days)
    time_from = timezone.now() - timedelta(days=past_days+1)

    events = models.MoonFrack.objects.visible_to(request.user)
    current_fracks = events.filter(
        arrival_time__gte=start_date,
        arrival_time__lt=timezone.now()).select_related(
            "moon_name",
            "moon_name__system",
            "moon_name__system__constellation",
            "moon_name__system__constellation__region",
    ).prefetch_related('frack',
                       "frack__ore",
                       "frack__ore__group"
                       )

    type_price = models.OrePrice.objects.filter(item_id=OuterRef('type_id'))

    output = {}
    str_ob_dict = {}

    for e in current_fracks:
        output[e.structure_id] = {
            "ObserverName": e.structure.location_name,
            "system": e.moon_name.system.name,
            "constellation": e.moon_name.system.constellation.name,
            "region": e.moon_name.system.constellation.region.name,
            "moon": {
                "name": e.moon_name.name,
                "id": e.moon_id
            },
            "extraction_end": e.arrival_time,
            "mined_ore": [],
            "total_m3": 0,
            "value": 0

        }
        for o in e.frack.all():
            if e.structure_id not in str_ob_dict:
                str_ob_dict[e.structure_id] = {}
            output[e.structure_id]['total_m3'] += o.total_m3
            str_ob_dict[e.structure_id][o.ore.name] = {
                "type": {
                    "id": o.ore_id,
                    "name": o.ore.name,
                    "cat": o.ore.group.name,
                    "cat_id": o.ore.group_id
                },
                "volume": 0,
                "total_volume": o.total_m3,
                "value": 0
            }

    observations = models.MiningObservation.objects \
        .filter(last_updated__gte=time_from) \
        .filter(observing_id__in=current_fracks.values_list("structure_id", flat=True)) \
        .values('structure', 'type_id') \
        .annotate(mined=(Sum('quantity') * F('type_name__volume'))) \
        .annotate(ore_value=ExpressionWrapper(
            Subquery(type_price.values('price')) * Sum('quantity'),
            output_field=FloatField())) \
        .annotate(name=F('type_name__name'))

    for o in observations:
        nme = o["name"].split(" ")[-1]
        if request.user.has_perm("moons.view_all"):
            str_ob_dict[o["structure"]][nme]["value"] += o["ore_value"]
        str_ob_dict[o["structure"]][nme]["volume"] += o['mined']

        if o['type_id'] in JACKPOT_IDS:
            output[o["structure"]]["jackpot"] = True

    for s, o in str_ob_dict.items():
        output[s]["mined_ore"] = list(o.values())

    return list(output.values())
"""


@api.get(
    "/extractions/past",
    response={200: List[schema.ExtractionEvent]},
    tags=["Observers"]
)
def get_moons_and_obs_past(request):
    if not request.user.has_perm("moons.view_all"):
        return []

    past_days = 30*12  # 12 months = max pull 6 events per moon

    return get_moons_and_extractions(request, past_days)


def get_moons_and_extractions(request, past_days):
    start_date = timezone.now() - timedelta(days=past_days)
    time_from = timezone.now() - timedelta(days=past_days+1)

    events = models.MoonFrack.objects.visible_to(request.user)
    current_fracks = events.filter(
        arrival_time__gte=start_date,
        arrival_time__lt=timezone.now()).select_related(
            "moon_name",
            "moon_name__system",
            "moon_name__system__constellation",
            "moon_name__system__constellation__region",
    ).prefetch_related('frack',
                       "frack__ore",
                       "frack__ore__group"
                       )

    type_price = models.OrePrice.objects.filter(item_id=OuterRef('type_id'))

    output = {}
    str_ob_dict = {}

    for e in current_fracks:
        output[e.id] = {
            "ObserverName": e.structure.location_name,
            "system": e.moon_name.system.name,
            "constellation": e.moon_name.system.constellation.name,
            "region": e.moon_name.system.constellation.region.name,
            "moon": {
                "name": e.moon_name.name,
                "id": e.moon_id
            },
            "extraction_end": e.arrival_time,
            "extraction_end": e.arrival_time,
            "mined_ore": [],
            "total_m3": 0,
            "value": 0,
            "structure_id": e.structure.location_id
        }

        for o in e.frack.all():
            if e.id not in str_ob_dict:
                str_ob_dict[e.id] = {}
            output[e.id]['total_m3'] += o.total_m3
            str_ob_dict[e.id][o.ore.name] = {
                "type": {
                    "id": o.ore_id,
                    "name": o.ore.name,
                    "cat": o.ore.group.name,
                    "cat_id": o.ore.group_id
                },
                "volume": 0,
                "total_volume": o.total_m3,
                "value": 0
            }

    observations = models.MiningObservation.objects \
        .filter(last_updated__gte=time_from) \
        .filter(observing_id__in=current_fracks.values_list("structure_id", flat=True)) \
        .values('structure', 'type_id', "last_updated") \
        .annotate(mined=(Sum('quantity') * F('type_name__volume'))) \
        .annotate(ore_value=ExpressionWrapper(
            Subquery(type_price.values('price')) * Sum('quantity'),
            output_field=FloatField())) \
        .annotate(name=F('type_name__name'))

    for o in observations:
        frack = what_frack_id(output, o)
        if frack == False:
            continue
        nme = o["name"].split(" ")[-1]
        if frack in str_ob_dict:
            if request.user.has_perm("moons.view_all"):
                str_ob_dict[frack][nme]["value"] += o["ore_value"]
                output[frack]['value'] += o["ore_value"]
            str_ob_dict[frack][nme]["volume"] += o['mined']

        if o['type_id'] in JACKPOT_IDS:
            output[frack]["jackpot"] = True

    for s, o in str_ob_dict.items():
        output[s]["mined_ore"] = list(o.values())

    return list(output.values())


@api.get(
    "/extractions/future",
    response={200: List[schema.ExtractionEvent]},
    tags=["Observers"]
)
def get_future_extractions(request):
    if not request.user.has_perm("moons.view_all"):
        return []

    start_date = timezone.now()

    events = models.MoonFrack.objects.visible_to(request.user)
    current_fracks = events.filter(
        arrival_time__gte=start_date).select_related(
            "moon_name",
            "moon_name__system",
            "moon_name__system__constellation",
            "moon_name__system__constellation__region",
    ).prefetch_related('frack',
                       "frack__ore",
                       "frack__ore__group"
                       )

    type_prices = OreHelper.get_ore_array_with_value()

    output = {}
    str_ob_dict = {}

    for e in current_fracks:
        output[e.structure_id] = {
            "ObserverName": e.structure.location_name,
            "system": e.moon_name.system.name,
            "constellation": e.moon_name.system.constellation.name,
            "region": e.moon_name.system.constellation.region.name,
            "moon": {
                "name": e.moon_name.name,
                "id": e.moon_id
            },
            "extraction_end": e.arrival_time,
            "mined_ore": [],
            "total_m3": 0,
            "value": 0
        }
        for o in e.frack.all():
            value = int(float(
                o.total_m3) / float(type_prices[o.ore_id]["volume"]) * float(type_prices[o.ore_id]["value"]))
            if e.structure_id not in str_ob_dict:
                str_ob_dict[e.structure_id] = {}
            output[e.structure_id]['total_m3'] += o.total_m3
            output[e.structure_id]['value'] += value

            str_ob_dict[e.structure_id][o.ore.name] = {
                "type": {
                    "id": o.ore_id,
                    "name": o.ore.name,
                    "cat": o.ore.group.name,
                    "cat_id": o.ore.group_id
                },
                "volume": 0,
                "total_volume": o.total_m3,
                "value": value
            }

    for s, o in str_ob_dict.items():
        output[s]["mined_ore"] = list(o.values())

    return list(output.values())


@api.get(
    "/moons/search",
    response={200: List[schema.IdName]},
    tags=["Search"]
)
def get_moon_search(request, search_text: str, limit: int = 10):
    return MapSystemMoon.objects.filter(name__icontains=search_text).values("name", id=F("moon_id"))[:limit]


@api.get(
    "/corporations/search",
    response={200: List[schema.Corporation]},
    tags=["Search"]
)
def get_corporation_search(request, search_text: str, limit: int = 10):
    return EveCorporationInfo.objects.filter(corporation_name__icontains=search_text)[:limit]


@api.get(
    "/characters/search",
    response={200: List[schema.Character]},
    tags=["Search"]
)
def get_character_search(request, search_text: str, limit: int = 10):
    return EveCharacter.objects.filter(character_name__icontains=search_text)[:limit]


@api.get(
    "/rental/list",
    response={200: List[schema.MoonRental]},
    tags=["Rentals"]
)
def get_moon_rentals(request):
    if not request.user.has_perm("moons.add_moonrental"):
        return []

    rentals = models.MoonRental.objects.filter(end_date__isnull=True).select_related(
        "moon", "moon__system", "contact", "corporation")
    out = []
    for r in rentals:
        out.append(
            {"moon": {
                "id": r.moon.moon_id,
                "name": r.moon.name
            },
                "system": {
                "id": r.moon.system.system_id,
                "name": r.moon.system.name
            },
                "contact": r.contact,
                "corporation": r.corporation,
                "price": r.price,
                "start_date": r.start_date
            }
        )

    return out


@api.get(
    "/rental/payments",
    response={200: List[schema.MoonRental]},
    tags=["Rentals"]
)
def get_moon_rental_payments(request):
    if not request.user.has_perm("moons.add_moonrental"):
        return []

    payments = Invoice.objects.filter(paid=False).select_related(
        "moon", "moon__system", "contact", "corporation")
    out = []
    for r in rentals:
        out.append(
            {"moon": {
                "id": r.moon.moon_id,
                "name": r.moon.name
            },
                "system": {
                "id": r.moon.system.system_id,
                "name": r.moon.system.name
            },
                "contact": r.contact,
                "corporation": r.corporation,
                "price": r.price,
                "start_date": r.start_date
            }
        )

    return out


@api.post(
    "/rental/new",
    response={200: schema.MoonRental, 403: str},
    tags=["Rentals"]
)
def post_moon_rental_new(request, rental: schema.NewMoonRental = Form(...)):
    if not request.user.has_perm("moons.add_moonrental"):
        return 403, "Permission Denied!"

    if models.MoonRental.objects.filter(moon_id=rental.moon_id, end_date__isnull=True).exists():
        return 403, "Moon Already Rented!"
    try:
        char = EveCharacter.objects.get(character_id=rental.contact_id)
    except EveCharacter.DoesNotExist:
        return 403, "Character Unknown to Auth"
    try:
        corp = EveCorporationInfo.objects.get(
            corporation_id=rental.corporation_id)
    except EveCorporationInfo.DoesNotExist:
        return 403, "Corporation Unknown to Auth"
    new_rental = models.MoonRental.objects.create(
        moon_id=rental.moon_id,
        contact=char,
        corporation=corp,
        price=rental.price,
        start_date=timezone.now()
    )
    return 200, {"moon": {
        "id": new_rental.moon.moon_id,
        "name": new_rental.moon.name
    },
        "system": {
        "id": new_rental.moon.system.system_id,
        "name": new_rental.moon.system.name
    },
        "contact": new_rental.contact,
        "corporation": new_rental.corporation,
        "price": new_rental.price,
        "start_date": new_rental.start_date
    }


@api.get(
    "/admin/list",
    response={200: List, 403: str},
    tags=["Admin"]
)
def get_corp_stats(request):
    if not request.user.is_superuser:
        return 403, "Permission Denied!"

    corps = CorporationAudit.objects.visible_to(request.user)

    output = []

    for c in corps:
        _c = {
            "name": c.corporation.corporation_name,
            "char_tokens": 0,
            "corp_tokens": 0,
            "obs": c.last_update_observers,
            "frack": "Never"
        }

        chars = CharacterAudit.objects.filter(character__corporation_id=c.corporation.corporation_id,
                                              characterroles__accountant=True,
                                              active=True)

        _c["char_tokens"] = chars.count()
        dt = chars.aggregate(Max("last_update_notif"))
        _c["frack"] = dt["last_update_notif__max"]
        scopes = ['esi-industry.read_corporation_mining.v1',
                  'esi-universe.read_structures.v1']

        tokens = Token.objects \
            .filter(character_id__in=chars.values_list("character__character_id", flat=True)) \
            .require_scopes(scopes)

        _c["corp_tokens"] = tokens.count()

        output.append(_c)

    return output
